import math


def convert_size(size):
    """ Size should be in bytes.
    Return a tuple (float_or_int_val, str_unit) """
    if size == 0:
        return (0, "B")
    KILOBYTE = 1024
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size, KILOBYTE)))
    p = math.pow(KILOBYTE, i)
    result = round(size/p, 2)
    return (result, size_name[i])


def badass_iso_8601_date_parser(date):
    # YYYY-MM-DDTHH:MM:SSZ
    date = date.rstrip("Z")
    date_part, time_part = date.split("T")
    months = ("January", "February", "March", "April", "May",
              "June", "July", "August", "September", "October",
              "November", "December")
    year, month, day = date_part.split("-")
    text = "{month} {day}, {year} at {time} UTC".format(month=months[int(month) - 1],
                                          day=day, year=year, time=time_part)
    return text


def dirty_metadata_parser(path):
    data = []
    with open(path, "r") as file:
        for line in file.readlines():
            key = ""
            value = ""
            parsing_key = True
            parsing_value = False
            cache = ""
            if line == "\n":
                break
            for char in line:
                if parsing_key and char == ":":
                    parsing_key = False
                    key = cache
                    cache = ""
                    continue
                if not parsing_key and not parsing_value:
                    parsing_value = True
                    continue
                if parsing_value and char == "\n":
                    value = cache
                    data.append((key, value))
                cache += char
    return data


from cyberpunk_theme import Cyberpunk
from cyberpunk_theme.widget import button, entry, listbox, label, frame
import tkstyle


def get_theme():
    theme = Cyberpunk()
    theme.add(get_button_style(), pattern="*Button")
    theme.add(get_entry_style(), pattern="*Entry")
    theme.add(get_listbox_style(), pattern="*Listbox")
    theme.add(get_frame_line_style(), pattern="*frame_line")
    theme.add(get_label_title_style(), pattern="*label_title")
    theme.add(get_label_notification_style(), pattern="*label_notification")
    theme.add(get_entry_owner_name_style(), pattern="*entry_owner_name")
    theme.add(get_entry_repo_name_style(), pattern="*entry_repo_name")
    theme.add(get_readonly_entry_style(), pattern="*info_toplevel*Entry")
    theme.add(get_readonly_entry_style(), pattern="*openlist_toplevel*Entry")
    theme.add(get_readonly_entry_style(), pattern="*error_toplevel*Entry")
    theme.add(get_readonly_entry_style(), pattern="*installer_toplevel*Entry")
    theme.add(get_readonly_entry_style(), pattern="*promoted_toplevel*Entry")
    return theme


def get_button_style():
    style = button.get_button_dark_filled_style()
    style.foreground = "gray"
    style.padX = 5
    style.padY = 0
    style.highlightThickness = 1
    style.highlightBackground = "#292D31"
    style.font = "Liberation Mono", 11, "normal"
    return style


def get_entry_style():
    style = entry.get_style()
    style.font = "Liberation Mono", 11, "normal"
    style.highlightThickness = 0
    return style


def get_listbox_style():
    style = listbox.get_style()
    style.font = "Liberation Mono", 11, "normal"
    return style


def get_frame_line_style():
    style = frame.get_style()
    style.background = "#009191"
    style.background = "#4E7F7A"
    return style


def get_label_title_style():
    style = label.get_style()
    style.foreground = "#009191"
    style.foreground = "#4E7F7A"
    style.font = ("Liberation Mono", 12, "bold")
    return style


def get_label_notification_style():
    style = label.get_style()
    style.foreground = "#606060"
    style.font = ("Liberation Mono", 9, "italic")
    return style


def get_entry_owner_name_style():
    style = entry.get_style()
    style.font = "Liberation Mono", 9, "normal"
    style.readonlyBackground = "#121519"
    return style


def get_entry_repo_name_style():
    style = entry.get_style()
    style.readonlyBackground = "#00312C"
    style.readonlyBackground = "#0B0E12"
    style.foreground = "#3CA46F"
    style.foreground = "#ABAEB2"
    style.font = "Liberation Mono", 11, "bold"
    return style


def get_highlight_style():
    style = tkstyle.Frame()
    style.highlightBackground = "#101010"
    #style.highlightColor = "#384848"
    style.highlightColor = "#384848"
    style.highlightBackground = "#384848"
    return style


def get_unhighlight_style():
    style = tkstyle.Frame()
    style.highlightThickness = 2
    style.highlightColor = "#121519"
    style.highlightBackground = "#121519"
    return style


def get_readonly_entry_style():
    style = tkstyle.Entry()
    style.readonlyBackground = "#121519"
    return style


def get_entry_description_style():
    style = tkstyle.Entry()
    style.font = "Liberation Mono", 11, "normal"
    return style


def get_info_entry_owner_repo_style():
    style = tkstyle.Entry()
    style.font = "Liberation Mono", 12, "bold"
    return style


def get_package_size_style():
    style = tkstyle.Entry()
    style.foreground = "#26718A"
    return style


def get_text_search_style():
    style = tkstyle.Entry()
    style.highlightThickness = 0
    style.foreground = "#009191"
    style.foreground = "gray"
    style.padX = 1
    style.padY = 0
    return style



import tkinter as tk
from viewable import Viewable
from megawidget.scrollbox import ScrollBox
from megawidget.toast import Toast
from hubstore.misc import funcs
from hubstore.misc.theme import get_entry_description_style, \
    get_info_entry_owner_repo_style, get_package_size_style
from cyberpunk_theme.widget.button import get_button_blue_filled_style, get_button_red_style


class Installer(Viewable):

    def __init__(self, parent_view, data=None):
        super().__init__()
        self._master = parent_view.body
        self._views = parent_view.views
        self._host = parent_view.host
        self._data = data
        self._strvar_description = tk.StringVar()
        self._strvar_owner_repo = tk.StringVar()
        self._strvar_tag_name = tk.StringVar()
        self._strvar_published_on = tk.StringVar()
        self._strvar_downloads_count = tk.StringVar(value="0")
        self._strvar_package_size = tk.StringVar(value="0 B")
        self._strvar_stargazers = tk.StringVar()
        self._repository_link = None
        self._image_cache = None
        self._asset = None
        self._setup()

    def _build(self):
        self._body = tk.Toplevel(name="installer_toplevel")
        self._body.resizable(False, False)
        #self._body.overrideredirect(1)
        self._body.title("Install an app")
        # install header
        self._install_header()
        # install central
        self._install_central()
        # install footer
        self._install_footer()

    def _on_map(self):
        super()._on_map()

    def _on_destroy(self):
        pass

    def _setup(self):
        if not self._data:
            return
        self._strvar_owner_repo.set(self._data["owner_repo"])
        self._strvar_description.set(self._data["description"])
        self._strvar_stargazers.set(self._data["stargazers"])
        self._strvar_tag_name.set(self._data["tag"])
        self._strvar_published_on.set(self._data["published_at"])
        self._repository_link = self._data["repository"]
        assets = self._data["assets"]
        if assets:
            self._asset = assets[0]
            self._strvar_downloads_count.set(self._asset["download_count"])
            cache = funcs.convert_size(self._asset["size"])
            self._strvar_package_size.set(cache)

    def _install_header(self):
        # install image
        canvas = tk.Canvas(self._body, width=400, height=100,
                           highlightthickness=0, borderwidth=0,
                           bg="#121519")
        canvas.pack(fill=tk.BOTH, pady=(0, 5))
        data = self._host.data.get_cover_image()
        image = tk.PhotoImage(data=data)
        self._image_cache = image
        canvas.create_image(0, 0, image=image, anchor="nw")
        # description
        description = tk.Entry(self._body,
                               textvariable=self._strvar_description,
                               state="readonly")
        description.pack(fill=tk.X, padx=0)
        get_entry_description_style().target(description)

    def _install_central(self):
        # install identification frame
        self._install_identification_frame()
        # install tag name
        self._install_tag_name()
        # install published_on
        self._install_published_on()
        # install stargazers_count
        self._install_stargazers_count()
        # install downloads_count
        self._install_downloads_count()

    def _install_identification_frame(self):
        # frame
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, padx=(0, 5), pady=5)
        # entry
        entry = tk.Entry(frame, textvariable=self._strvar_owner_repo,
                         state="readonly")
        entry.pack(side=tk.LEFT, fill=tk.X, expand=1)
        get_info_entry_owner_repo_style().target(entry)
        # button
        command = lambda self=self: self._host.open_website(self._repository_link)
        button = tk.Button(frame, text="Repository",
                           command=command)
        button.pack(side=tk.RIGHT)

    def _install_tag_name(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, pady=(0, 5))
        label = tk.Label(frame, text="Tag:")
        label.pack(side=tk.LEFT)
        entry = tk.Entry(frame, textvariable=self._strvar_tag_name,
                         state="readonly")
        entry.pack(side=tk.LEFT, fill=tk.X, expand=1)
        #entry.pack(anchor="w", fill=tk.X, pady=(0, 5))

    def _install_published_on(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, pady=(10, 0))
        label = tk.Label(frame, text="Published:")
        label.pack(side=tk.LEFT)
        entry = tk.Entry(frame, textvariable=self._strvar_published_on,
                         state="readonly")
        entry.pack(side=tk.LEFT, fill=tk.X, expand=1)
        #entry.pack(anchor="w", fill=tk.X, pady=(10, 5))

    def _install_stargazers_count(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, pady=(5, 5))
        label = tk.Label(frame, text="Stargazers:")
        label.pack(side=tk.LEFT)
        entry = tk.Entry(frame, textvariable=self._strvar_stargazers,
                         state="readonly")
        entry.pack(side=tk.LEFT, expand=1, fill=tk.X)

    def _install_downloads_count(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, pady=(0, 5))
        label = tk.Label(frame, text="Downloads:")
        label.pack(side=tk.LEFT)
        entry = tk.Entry(frame, textvariable=self._strvar_downloads_count,
                         state="readonly")
        entry.pack(side=tk.LEFT, expand=1, fill=tk.X)
        #entry.pack(anchor="w", fill=tk.X, pady=(0, 5))

    def _install_footer(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, padx=3, pady=(20, 3))
        # button download
        button_install = tk.Button(frame, text="Install",
                                   command=self._on_click_install)
        button_install.pack(side=tk.RIGHT, padx=(3, 0))
        get_button_style_4().target(button_install)
        # button cancel
        button_cancel = tk.Button(frame, text="Cancel",
                                  command=self.destroy)
        button_cancel.pack(side=tk.RIGHT)
        # package
        entry = tk.Entry(frame, textvariable=self._strvar_package_size,
                         state="readonly")
        entry.pack(side=tk.LEFT, fill=tk.X, anchor="w")
        get_package_size_style().target(entry)

    def _on_click_install(self):
        if not self._asset:
            Toast(self._body, message="No asset to download")
            return
        self.destroy()
        owner, repo = self._data["owner_repo"].split("/")
        name = self._asset["name"]
        url = self._asset["url"]
        self._host.install(owner, repo, name, url)



import tkinter as tk
from tkinter.scrolledtext import ScrolledText
from viewable import Viewable
from megawidget.scrollbox import ScrollBox
from megawidget.toast import Toast
from hubstore.misc.theme import get_entry_description_style, \
    get_info_entry_owner_repo_style, get_package_size_style
from cyberpunk_theme.widget.button import get_button_blue_filled_style, get_button_red_style


class Report(Viewable):

    def __init__(self, parent_view, data=None):
        super().__init__()
        self._master = parent_view.body
        self._views = parent_view.views
        self._host = parent_view.host
        self._data = data
        self._strvar_description = tk.StringVar(value="Oops, an error occurred !")
        self._strvar_owner_repo = tk.StringVar()
        self._text_widget = None
        self._image_cache = None
        self._scrollbox = None
        self._setup()

    def _build(self):
        self._body = tk.Toplevel(name="error_toplevel")
        self._body.resizable(False, False)
        self._body.title("Crash report")
        #self._body.geometry("350x250")
        #self._body.overrideredirect(1)
        # install header
        self._install_header()
        # install central
        self._install_central()
        # install footer
        self._install_footer()

    def _on_map(self):
        super()._on_map()
        self._strvar_owner_repo.set(self._data["owner_repo"])
        self._text_widget.insert("1.0", self._data["error"])
        self._text_widget.config(state="disabled")

    def _on_destroy(self):
        pass

    def _install_header(self):
        """
        # install image
        canvas = tk.Canvas(self._body, width=400, height=100,
                           highlightthickness=0, borderwidth=0,
                           bg="#121519")
        canvas.pack(fill=tk.BOTH, pady=(0,5))
        with open("/home/alex/default_info_cover.png",
                  "rb") as file:
            data = file.read()
        image = tk.PhotoImage(data=data)
        self._image_cache = image
        canvas.create_image(0, 0, image=image, anchor="nw")
        """
        # description
        description = tk.Label(self._body,
                               textvariable=self._strvar_description)
        description.pack(pady=(10, 10), anchor="center")
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, expand=1)
        label = tk.Label(frame, text="Application:")
        label.pack(side=tk.LEFT)
        entry = tk.Entry(frame, textvariable=self._strvar_owner_repo,
                         state="readonly")
        entry.pack(side=tk.LEFT, expand=1, fill=tk.X)
        #get_entry_description_style().target(description)

    def _install_central(self):
        # install identification frame
        self._text_widget = ScrolledText(self._body, width=52, height=13,
                                         wrap="word")
        self._text_widget.pack(pady=10)

    def _install_footer(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, padx=3, pady=(15, 3))
        # button report
        button_report = tk.Button(frame, text="Report",
                                  command=self._on_click_report)
        button_report.pack(side=tk.RIGHT)
        get_button_blue_filled_style().target(button_report)
        # button copy
        button_copy = tk.Button(frame, text="Copy",
                                command=self._on_click_copy)
        button_copy.pack(side=tk.RIGHT, padx=(0, 3))
        # button close
        button_close = tk.Button(frame, text="Close",
                                 command=self.destroy)
        button_close.pack(side=tk.RIGHT, padx=(0, 3))

    def _setup(self):
        """
        self._data =  {"error": error, "owner_repo": owner_repo,
                    "repository": repository}
        """
        pass

    def _on_click_copy(self):
        self._body.clipboard_clear()
        text = self._data["error"]
        self._body.clipboard_append(text)
        self._body.update()
        Toast(self._body, message="Traceback copied !", duration=700)

    def _on_click_report(self):
        self._host.open_website(self._data["repository"])



import tkinter as tk
from viewable import Viewable
from megawidget.scrollbox import ScrollBox
from megawidget.toast import Toast
from hubstore.misc.theme import get_entry_description_style
from hubstore.misc.theme import get_info_entry_owner_repo_style
from cyberpunk_theme.widget.button import get_button_red_style, get_button_blue_filled_style


class Info(Viewable):

    def __init__(self, parent_view, data=None):
        super().__init__()
        self._data = data
        self._master = parent_view.body
        self._views = parent_view.views
        self._host = parent_view.host
        self._strvar_description = tk.StringVar()
        self._strvar_owner_repo = tk.StringVar()
        self._strvar_version = tk.StringVar()
        self._strvar_updated_since = tk.StringVar()
        self._strvar_rollback_to = tk.StringVar()
        self._intvar_favorite = tk.IntVar(value=0)
        self._image_cache = None
        self._setup()

    def _build(self):
        self._body = tk.Toplevel(name="info_toplevel")
        self._body.title("App information")
        self._body.resizable(False, False)
        #self._body.overrideredirect(1)
        # install header
        self._install_header()
        # install central
        self._install_central()
        # install footer
        self._install_footer()

    def _on_map(self):
        super()._on_map()

    def _on_destroy(self):
        pass

    def _install_header(self):
        # install image
        canvas = tk.Canvas(self._body, width=400, height=100,
                           highlightthickness=0, borderwidth=0,
                           bg="#121519")
        canvas.pack(fill=tk.BOTH, pady=(0,5))
        data = self._host.data.get_cover_image()
        image = tk.PhotoImage(data=data)
        self._image_cache = image
        canvas.create_image(0, 0, image=image, anchor="nw")
        # description
        description = tk.Entry(self._body,
                               textvariable=self._strvar_description,
                               state="readonly")
        description.pack(fill=tk.X, padx=2)
        get_entry_description_style().target(description)

    def _install_central(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.BOTH, expand=1, padx=(2, 10), pady=(10, 30))
        frame.columnconfigure(0, weight=1)
        frame.rowconfigure(0, pad=5)
        frame.rowconfigure(1, pad=5)
        frame.rowconfigure(2, pad=5)
        frame.rowconfigure(3, pad=5)
        # install identification frame
        self._install_identification_frame(frame)
        # install version text
        self._install_version_text(frame)
        # install updating frame
        self._install_updating_frame(frame)
        # install rollback frame
        if self._data["backup_version"]:
            self._install_rollback_frame(frame)

    def _install_identification_frame(self, frame):
        # entry
        entry = tk.Entry(frame, textvariable=self._strvar_owner_repo,
                         state="readonly")
        entry.grid(row=0, column=0, sticky="we")
        get_info_entry_owner_repo_style().target(entry)
        # button
        command = (lambda self=self:
                   self._host.open_website(self._data["repository"]))
        button = tk.Button(frame, text="Repository",
                           command=command)
        button.grid(row=0, column=1, sticky="we")

    def _install_version_text(self, frame):
        subframe = tk.Frame(frame)
        subframe.grid(row=1, column=0, sticky="we")
        label = tk.Label(subframe, text="Version:")
        label.pack(side=tk.LEFT)
        entry = tk.Entry(subframe, textvariable=self._strvar_version,
                         state="readonly")
        entry.pack(side=tk.LEFT, expand=1, fill=tk.X)
        #entry.grid(row=1, column=0, sticky="we")

    def _install_updating_frame(self, frame):
        # entry
        entry = tk.Entry(frame, textvariable=self._strvar_updated_since,
                         state="readonly")
        entry.grid(row=2, column=0, sticky="we")
        # button
        button = tk.Button(frame, text="Update",
                           command=self._on_click_update)
        button.grid(row=2, column=1, sticky="we")

    def _install_rollback_frame(self, frame):
        # entry
        entry = tk.Entry(frame, textvariable=self._strvar_rollback_to,
                         state="readonly")
        entry.grid(row=3, column=0, sticky="we")
        # button
        button = tk.Button(frame, text="Rollback",
                           command=self._on_click_rollback)
        button.grid(row=3, column=1, sticky="we")

    def _install_footer(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, padx=3, pady=3)
        # button uninstall
        button_uninstall = tk.Button(frame, text="Uninstall",
                                     command=self._on_click_uninstall)
        button_uninstall.pack(side=tk.LEFT)
        get_button_red_style().target(button_uninstall)
        # button run
        button_run = tk.Button(frame, text="Run",
                                 command=self._on_click_run)
        button_run.pack(side=tk.RIGHT, padx=(3, 0))
        get_button_blue_filled_style().target(button_run)
        # button close
        button_close = tk.Button(frame, text="Close",
                                 command=self.destroy)
        button_close.pack(side=tk.RIGHT, padx=(10, 0))
        # checkbox Favorite
        checkbox = tk.Checkbutton(frame, text="Favorite",
                                  variable=self._intvar_favorite,
                                  onvalue=1, offvalue=0,
                                  command=self._on_click_favorite)
        checkbox.pack(side=tk.RIGHT, anchor="s")

    def _setup(self):
        # populate
        self._strvar_description.set(self._data["description"])
        self._strvar_owner_repo.set(self._data["owner_repo"])
        self._strvar_version.set(self._data["version"])
        if self._data["favorite"]:
            self._intvar_favorite.set(1)
        backup_version = self._data["backup_version"]
        if backup_version:
            cache = "Rollback to {}".format(backup_version)
            self._strvar_rollback_to.set(cache)
        updated_since = self._data["updated_since"]
        if updated_since >= 1000:
            cache = "Updated since 999+ days ago"
        elif updated_since > 1:
            cache = "Updated since {} days ago".format(updated_since)
        else:
            cache = "Updated recently"
        self._strvar_updated_since.set(cache)

    def _on_click_favorite(self):
        val = self._intvar_favorite.get()
        val = True if val == 1 else False
        self._host.favorite(self._data["owner_repo"], val=val)

    def _on_click_update(self):
        owner_repo = self._data["owner_repo"]
        owner, repo = owner_repo.split("/")
        self.destroy()
        self._host.update(owner, repo)

    def _on_click_rollback(self):
        owner_repo = self._data["owner_repo"]
        owner, repo = owner_repo.split("/")
        self.destroy()
        self._views.main.show_toast("Processing Rollback...", duration=None)
        self._host.rollback(owner, repo)

    def _on_click_run(self):
        owner_repo = self._data["owner_repo"]
        owner, repo = owner_repo.split("/")
        self.destroy()
        self._host.run_app(owner, repo)

    def _on_click_uninstall(self):
        owner_repo = self._data["owner_repo"]
        owner, repo = owner_repo.split("/")
        self.destroy()
        self._views.main.show_toast("Uninstalling...", duration=None)
        self._host.uninstall(owner, repo)



import tkinter as tk
from viewable import Viewable
from megawidget.pathfield import PathField
import os
import os.path
from cyberpunk_theme.widget.button import get_button_blue_filled_style, get_button_yellow_style


ABOUT_TEXT = """Download, install, manage, promote, and run Python desktop apps with Hubstore.

Hubstore is part of the Pyrustic Open Ecosystem.

"""


WEBSITE = "https://pyrustic.github.io"


class About(Viewable):
    def __init__(self, parent_view):
        super().__init__()
        self._app = parent_view.app
        self._master = parent_view.body
        self._host = parent_view.host
        self._strvar_apps_directory = tk.StringVar(value="")

    @property
    def directory_path(self):
        return self._strvar_apps_directory.get()

    def _build(self):
        self._body = tk.Toplevel(self._master)
        self._body.title("About Hubstore")
        self._body.resizable(False, False)
        # install header
        self._install_header()
        # install central
        self._install_central()
        # install directory form
        self._install_directory_form()
        # install footer
        self._install_footer()

    def _on_map(self):
        super()._on_map()
        apps_directory = self._host.data.get_apps_directory()
        self._strvar_apps_directory.set(apps_directory)

    def _on_destroy(self):
        pass

    def _install_header(self):
        # install image
        canvas = tk.Canvas(self._body, width=400, height=100,
                           highlightthickness=0, borderwidth=0,
                           bg="#121519")
        canvas.pack(fill=tk.BOTH, pady=(0,5))
        data = self._host.data.get_cover_image()
        image = tk.PhotoImage(data=data)
        self._image_cache = image
        canvas.create_image(0, 0, image=image, anchor="nw")
        label = tk.Label(self._body, text="H U B S T O R E", fg="#4E7F7A",
                         font=("Liberation Mono", 25, "bold"))
        label.pack(anchor="center")

    def _install_central(self):
        # text
        text = tk.Text(self._body, width=1, height=6, bg="#121519",
                       fg="gray", font=("Liberation Mono", 12, "normal"),
                       wrap="word", borderwidth=0, highlightthickness=0)
        text.insert("1.0", ABOUT_TEXT)
        text.config(state="disabled")
        text.pack(fill=tk.X, pady=10, padx=3, expand=1)

    def _install_directory_form(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, padx=3, pady=(0, 20))
        label = tk.Label(frame, text="Directory to store apps", padx=0, borderwidth=0)
        label.pack(anchor="w")
        pathentry = PathField(frame, browse="dir",
                              initialdir=os.path.expanduser("~"),
                              textvariable=self._strvar_apps_directory)
        pathentry.pack(fill=tk.X, anchor="w")

    def _install_footer(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, padx=3, pady=3)
        command = lambda self=self: self._on_click_website()
        button_website = tk.Button(frame, text="Website", command=command)
        button_website.pack(side=tk.LEFT, padx=(0, 3))
        get_button_yellow_style().target(button_website)
        command = lambda self=self: self._on_click_accept()
        button_accept = tk.Button(frame, text="Accept", command=command)
        button_accept.pack(side=tk.RIGHT)
        get_button_blue_filled_style().target(button_accept)
        command = lambda self=self: self.destroy()
        button_close = tk.Button(frame, text="Close", command=command)
        button_close.pack(side=tk.RIGHT, padx=3)

    def _on_click_accept(self):
        self._host.submit_apps_directory(self._strvar_apps_directory.get())
        self.destroy()

    def _on_click_website(self):
        self._host.open_website(WEBSITE)



import tkinter as tk
from viewable import Viewable
from megawidget.scrollbox import ScrollBox
from megawidget.toast import Toast
from hubstore.misc.theme import get_entry_description_style, \
    get_info_entry_owner_repo_style, get_package_size_style
from cyberpunk_theme.widget.button import get_button_blue_filled_style, get_button_red_style


class Openlist(Viewable):

    def __init__(self, parent_view, data=None):
        super().__init__()
        self._data = data
        self._master = parent_view.body
        self._views = parent_view.views
        self._host = parent_view.host
        self._strvar_description = tk.StringVar(value="Currently running")
        self._strvar_owner_repo = tk.StringVar(value="pyrustic/demo")
        self._strvar_tag_name = tk.StringVar(value="v10.0.2")
        self._strvar_published_on = tk.StringVar(value="January 21, 2021 at 12:20 UTC")
        self._strvar_created_on = tk.StringVar(value="March 14, 2021 at 09:10 UTC")
        self._strvar_stargazers_count = tk.StringVar(value="34")
        self._strvar_downloads_count = tk.StringVar(value="345")
        self._strvar_package_size = tk.StringVar(value="24 KB")
        self._image_cache = None
        self._scrollbox = None

    def _build(self):
        self._body = tk.Toplevel(name="openlist_toplevel")
        self._body.geometry("350x250")
        self._body.title("Apps Manager")
        self._body.resizable(False, False)
        #self._body.overrideredirect(1)
        self._scrollbox = ScrollBox(self._body)
        self._scrollbox.pack(expand=1, fill=tk.BOTH)
        # install header
        self._install_header()
        # install central
        self._install_central()
        # install footer
        self._install_footer()

    def _on_map(self):
        super()._on_map()

    def _on_destroy(self):
        pass

    def _install_header(self):
        """
        # install image
        canvas = tk.Canvas(self._body, width=400, height=100,
                           highlightthickness=0, borderwidth=0,
                           bg="#121519")
        canvas.pack(fill=tk.BOTH, pady=(0,5))
        with open("/home/alex/default_info_cover.png",
                  "rb") as file:
            data = file.read()
        image = tk.PhotoImage(data=data)
        self._image_cache = image
        canvas.create_image(0, 0, image=image, anchor="nw")
        """
        # description
        description = tk.Label(self._scrollbox.box, font=("Liberation Mono", 11, "bold"),
                               textvariable=self._strvar_description)
        description.pack(pady=(10, 10), anchor="center")
        #get_entry_description_style().target(description)

    def _install_central(self):
        # install identification frame
        for item in self._data:
            owner = item["owner"]
            repo = item["repo"]
            process_id = item["id"]
            self._install_identification_frame(owner, repo, process_id)

    def _install_identification_frame(self, owner, repo, process_id):
        # frame
        frame = tk.Frame(self._scrollbox.box)
        frame.pack(fill=tk.X, padx=(3, 5), pady=5)
        # entry
        owner_repo = "{}/{}".format(owner, repo)
        strvar = tk.StringVar(value=owner_repo)
        entry = tk.Entry(frame, textvariable=strvar,
                         state="readonly")
        entry.pack(side=tk.LEFT, fill=tk.X, expand=1)
        #get_info_entry_owner_repo_style().target(entry)
        # button
        command = (lambda self=self, frame=frame,
                          process_id=process_id:
                   self._on_close_app(frame, process_id))
        button = tk.Button(frame, text="Stop",
                           command=command)
        button.pack(side=tk.LEFT, padx=(5, 0))
        get_button_red_style().target(button)

    def _install_footer(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, padx=3, pady=(15, 3))
        # button close
        button = tk.Button(frame, text="Close",
                           command=self.destroy)
        button.pack(side=tk.RIGHT)

    def _on_close_app(self, frame, process_id):
        self._host.close_app(process_id)
        frame.destroy()



import tkinter as tk
from viewable import Viewable
from threadom import Threadom
from megawidget.toast import Toast
from hubstore.host import Host
from hubstore.view.views import Views
from hubstore.view.toolbar import Toolbar
from hubstore.view.pane import Pane
from hubstore.view.about import About
from hubstore.view.installer import Installer
from hubstore.view.info import Info
from hubstore.view.report import Report
from hubstore.view.openlist import Openlist
from hubstore.view.promoted import Promoted


class Main(Viewable):

    def __init__(self, app):
        super().__init__()
        self._app = app
        self._master = app.root
        self._threadom = Threadom(app.root)
        self._views = Views()
        self._host = Host(self)
        self._views.main = self
        self._toast = None

    @property
    def app(self):
        return self._app

    @property
    def host(self):
        return self._host

    @property
    def views(self):
        return self._views

    def show_toast(self, message, duration=1234):
        self.close_toast()
        toast = Toast(self._app.root, message=message,
                      duration=duration)
        toast.update_idletasks()
        if duration is None:
            self._toast = toast

    def show_notification(self, message):
        self._views.pane.set_notification(message)

    def close_toast(self):
        if self._toast:
            self._toast.destroy()
            self._toast = None

    def clear_notification(self):
        self._views.pane.set_notification("")

    def open_toplevel_about(self):
        about = About(self)
        about.build_wait()

    def open_toplevel_installer(self, data):
        installer = Installer(self, data=data)
        installer.build_wait()

    def open_toplevel_report(self, data):
        report = Report(self, data=data)
        report.build_wait()

    def open_toplevel_info(self, data):
        info = Info(self, data=data)
        info.build_wait()

    def open_toplevel_promoted(self, data):
        promoted = Promoted(self, data=data)
        promoted.build_wait()

    def open_toplevel_openlist(self, data):
        openlist = Openlist(self, data=data)
        openlist.build_wait()

    def _build(self):
        self._body = tk.Frame(self._master)
        # install toolbar
        self._install_toolbar()
        # install pane
        self._install_pane()

    def _on_map(self):
        if self._host.on_start_app():
            # binding
            self._binding()
            self._views.toolbar.wake_search_entry()
        #from hubstore.view.downloader import Downloader
        #from hubstore.view.info import Info
        #Info(self).build()
        #Downloader(self).build()

    def _install_toolbar(self):
        toolbar = Toolbar(self._body, self._views, self._host)
        toolbar.build_pack(fill=tk.X, padx=2, pady=2)
        self._views.toolbar = toolbar

    def _install_pane(self):
        pane = Pane(self)
        pane.build_pack(expand=1, fill=tk.BOTH)
        self._views.pane = pane

    def _binding(self):
        self._body.focus_set()
        command = lambda e, views=self._views: views.toolbar.wake_search_entry()
        self._body.bind("<Control-f>", command)
        self._body.bind("<Control-F>", command)



import os
import os.path
import sys
import time
import webbrowser
import subprocess
import pkgutil
from threadom import Threadom
from shared import Jason
from hubstore import core
from hubstore.misc import constant


class Host:
    def __init__(self, main_view):
        self._main_view = main_view
        self._app = main_view.app
        self._root = main_view.app.root
        self._views = main_view.views
        self._data = Data(self)
        self._initialized = None
        self._threadom = None
        self._quota = None
        self._kurl = None
        self._count_processes = 0
        self._processes = list()
        self._setup()

    @property
    def initialized(self):
        return self._initialized

    @property
    def threadom(self):
        return self._threadom

    @property
    def data(self):
        return self._data

    def on_start_app(self):
        if not self.initialized:
            self._main_view.open_toplevel_about()
        if not self.initialized:
            self._app.exit()
            return False
        apps_list = self._data.get_favorites_apps()
        title = "Favorite Apps ({})".format(len(apps_list))
        if not apps_list:
            apps_list = self._data.get_all_apps()
            title = "All Apps ({})".format(len(apps_list))
        if apps_list:
            self._views.pane.populate(title, apps_list)
            self._views.toolbar.update_apps_list(self._data.get_all_apps())
        return True

    def submit_apps_directory(self, path):
        self._initialized = False
        self._store = None
        data = core.init_hubstore(path)
        error_code = data["error_code"]
        error = data["error"]
        if error_code == 0:
            self._initialized = True
        elif error_code == 1:
            self._main_view.show_toast("Failed to create 'hubstore-apps' folder")
        elif error_code == 2:
            self._main_view.show_toast("Failed to register Hubstore in $HOME/PyrusticData")
        else:
            self._main_view.show_toast("Unknown error code")

    def open_website(self, url):
        command = lambda: webbrowser.open(url, new=2)
        self._root.after(1, command)

    def search(self, owner, repo):
        path = core.get_path(owner, repo)
        if path:
            self.show_app_info(owner, repo)
            #self._search_offline(owner, repo)
        else:
            self._search_online(owner, repo)

    def update(self, owner, repo):
        self._search_online(owner, repo)

    def rollback(self, owner, repo):
        is_success, error = core.rollback(owner, repo)
        if is_success:
            message = "Rollback done with success !"
            store = core.get_store()
            owner_repo = "{}/{}".format(owner, repo)
            data = store.get(owner_repo)
            data["backup_version"] = None
            data.save()
        else:
            message = "Failed to perform rollback"
        self._main_view.show_toast(message)

    def run_app(self, owner, repo):
        self._count_processes += 1
        process_id = self._count_processes
        self._processes.append({"id": process_id,
                                "owner": owner,
                                "repo": repo,
                                "process": None})
        target = (lambda self=self, owner=owner,
                          repo=repo, process_id=process_id:
                  self._run(owner, repo, process_id))
        consumer = (lambda error, self=self,
                           process_id=process_id:
                    self._on_app_exit(error, process_id))
        apps_open = len(self._processes)
        self._views.toolbar.update_open_count(apps_open)
        self._threadom.run(target, consumer=consumer)

    def close_app(self, process_id):
        try:
            i, data = self._get_process(process_id)
            data["process"].terminate()
        except Exception as e:
            pass

    def show_apps_from(self, owner):
        apps = self._data.get_all_apps()
        cache = []
        for x, repo in apps:
            if x == owner:
                cache.append((owner, repo))
        self._views.pane.populate("Apps from '{}' ({})".format(owner, len(cache)),
                                  cache)

    def show_app_info(self, owner, repo):
        owner_repo = "{}/{}".format(owner, repo)
        store = core.get_store()
        data = store.get(owner_repo)
        data = data.copy()
        data["owner_repo"] = owner_repo
        data["favorite"] = True if owner_repo in store.get("favorites") else False
        data["repository"] = "https://github.com/{}".format(owner_repo)
        current_time = int(time.time())
        timestamp_install = data["timestamp_install"]
        cache = current_time - timestamp_install
        cache = cache / (60*60*24)
        data["updated_since"] = int(cache)
        self._main_view.open_toplevel_info(data)

    def show_all_apps(self):
        apps_list = self._data.get_all_apps()
        title = "All Apps ({})".format(len(apps_list))
        if not apps_list:
            title = ""
        self._views.pane.populate(title, apps_list)
        self._views.toolbar.update_apps_list(self._data.get_all_apps())

    def show_favorite_apps(self):
        apps_list = self._data.get_favorites_apps()
        title = "Favorite Apps ({})".format(len(apps_list))
        if not apps_list:
            title = ""
        self._views.pane.populate(title, apps_list)

    def show_promoted(self):
        target = lambda self=self: self._show_promoted()
        consumer = (lambda data, self=self:
                    self._main_view.open_toplevel_promoted(data))
        self._threadom.run(target, consumer=consumer)

    def show_open(self):
        if not self._processes:
            self._main_view.show_toast("No app is running")
            return
        cache = []
        for item in self._processes:
            owner = item["owner"]
            repo = item["repo"]
            process_id = item["id"]
            cache.append({"owner": owner, "repo": repo,
                          "id": process_id})
        cache.reverse()
        self._main_view.open_toplevel_openlist(cache)

    def show_about(self):
        self._main_view.open_toplevel_about()

    def install(self, owner, repo, name, url):
        message = "Installation..."
        self._main_view.show_toast(message, duration=None)
        if core.get_path(owner, repo):
            is_success, error = core.backup(owner, repo)
            if not is_success:
                self._main_view.show_toast("Failed to make a backup !")
                return
        # download
        is_success, error, tempfile = core.download(name, url, self._kurl)
        if not is_success:
            self._main_view.show_toast("Failed to download the asset !")
            return
        # install
        is_success, error = core.install(owner, repo, name, tempfile)
        if not is_success:
            self._main_view.show_toast("Failed to install the asset !")
            return
        # update app information
        self._update_app_info(owner, repo)
        # update the list of apps
        message = "Successfully installed '{}/{}' !".format(owner, repo)
        self._main_view.show_toast(message)
        apps_list = self._data.get_all_apps()
        title = "All Apps ({})".format(len(apps_list))
        if apps_list:
            self._views.pane.populate(title, apps_list)
            self._views.toolbar.update_apps_list(self._data.get_all_apps())

    def uninstall(self, owner, repo):
        is_success, error = core.uninstall(owner, repo)
        message = "Successfully uninstalled !"
        if not is_success:
            message = "Failed to uninstall"
        self._main_view.show_toast(message)
        self.show_all_apps()

    def favorite(self, owner_repo, val=True):
        store = core.get_store()
        data = store.get("favorites")
        if val:
            if owner_repo not in data:
                data.append(owner_repo)
        else:
            cache = None
            for i, item in enumerate(data):
                if item == owner_repo:
                    cache = i
                    break
            if cache is not None:
                del data[cache]
        data.save()

    # ======
    def _setup(self):
        self._initialized = True
        if core.should_init_hubstore():
            self._initialized = False
        self._kurl = core.get_kurl()
        self._threadom = Threadom(self._root)

    def _search_offline(self, owner, repo):
        print("search offline 2")
        print("SEARCH OFFLINE: ", owner, repo)

    def _search_online(self, owner, repo):
        self._main_view.show_toast("Fetching...", duration=None)
        # update quota
        updated, status_text = self._update_quota()
        if not updated:
            self._main_view.show_toast("Check your connection")
            return
        # check rate
        if self._quota - 2 == 0:
            message = "You are reaching the quota limit !\nPlease try in a few minutes"
            self._main_view.show_toast(message)
            return
        # open download pane
        cache = core.fetch(owner, repo, self._kurl)
        code, error, data = cache
        self._main_view.close_toast()
        if code in (200, 304):
            self._quota -= 2
            self._main_view.open_toplevel_installer(data)
        else:
            self._main_view.show_toast(error)

    def _update_quota(self):
        if self._quota:
            return True, None
        status_code, status_text, data = core.rate(self._kurl)
        updated = False
        if status_code in (200, 304):
            self._quota = data["remaining"]
            updated = True
        return updated, status_text

    def _update_app_info(self, owner, repo):
        store = core.get_store()
        metadata = core.app_metadata(owner, repo)
        owner_repo = "{}/{}".format(owner, repo)
        info = store.get(owner_repo)
        info["version"] = metadata["version"]
        info["description"] = metadata["description"]
        info["small_img"] = metadata["small_img"]
        info["large_img"] = metadata["large_img"]
        info.save()

    def _run(self, owner, repo, process_id):
        store = core.get_store()
        app_pkg = core.get_app_pkg(owner, repo, store=store)
        path = core.get_path(owner, repo, store=store)
        try:
            p = subprocess.Popen([sys.executable, "-m", app_pkg],
                                 cwd=path, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            i, info = self._get_process(process_id)
            self._processes[i]["process"] = p
            out, err = p.communicate()
        except Exception as e:
            err = e
        return err

    def _on_app_exit(self, error, process_id):
        i, _ = self._get_process(process_id)
        if error:
            owner = self._processes[i]["owner"]
            repo = self._processes[i]["repo"]
            owner_repo = "{}/{}".format(owner, repo)
            repository = "https://github.com/{}/{}/issues".format(owner, repo)
            data = {"error": error, "owner_repo": owner_repo,
                    "repository": repository}
            self._main_view.open_toplevel_report(data)
        del self._processes[i]
        apps_open = len(self._processes)
        self._views.toolbar.update_open_count(apps_open)

    def _get_process(self, process_id):
        for i, item in enumerate(self._processes):
            if item["id"] == process_id:
                return i, item

    def _show_promoted(self):  # TODO improve + refactor this in a class
        # generate list of promoted apps
        store = core.get_store()
        apps = store.get("apps")
        cache = []
        owners_repos = []
        for owner, repos in apps.items():
            for repo in repos:
                habitat_path = core.get_path(owner, repo)
                app_pkg = core.get_app_pkg(owner, repo)
                location = os.path.join(habitat_path, app_pkg,
                                        "pyrustic_data",
                                        "hubstore")
                jason = Jason("promotion.json", location=location,
                              readonly=True)
                if not jason.data:
                    continue
                for owner_repo, description in jason.data.items():
                    if owner_repo in owners_repos:
                        continue
                    owners_repos.append(owner_repo)
                    cache.append({"repository": owner_repo,
                                  "description": description})
        #random.choice(cache)
        return cache


class Data:

    def __init__(self, host):
        self._host = host
        self._store = None

    @property
    def store(self):
        if not self._store:
            self._store = core.get_store()
        return self._store

    def get_apps_directory(self):
        if not self._host.initialized:
            return os.path.expanduser("~")
        location = os.path.join(constant.PYRUSTIC_DATA, "hubstore")
        jason = Jason("meta.json", location=location)
        return os.path.dirname(jason.data.get("hubstore-apps"))

    def get_all_apps(self):
        store = core.get_store()
        apps = store.get("apps")
        cache = []
        for owner, repos in apps.items():
            for repo in repos:
                cache.append((owner, repo))
        cache.sort(key=lambda x: x[1])
        return cache

        #return (("facebook", "whatsapp"), ("facebook", "instagram"),
        #        ("google", "search"), ("google", "android"),
        #        ("google", "chromium"))

    def get_favorites_apps(self):
        store = core.get_store()
        apps = store.get("favorites")
        cache = []
        for app in apps:
            cache.append(app.split("/"))
        return cache

    def get_image(self, owner, repo):
        store = core.get_store()
        owner_repo = "{}/{}".format(owner, repo)
        data = store.get(owner_repo)
        app_pkg = data["app_pkg"]
        small_img = data["small_img"]
        if not small_img:
            return self.get_default_image()
        habitat = core.get_path(owner, repo)
        cache = os.path.join(habitat, app_pkg)
        path = core.normpath(cache, small_img)
        if os.path.isfile(path):
            with open(path, "rb") as file:
                data = file.read()
            return data
        return self.get_default_image()

    def get_cover_image(self):
        return self.get_default_cover_image()

    def get_default_image(self):
        data = pkgutil.get_data("hubstore",
                                "asset/default_hubstore_tile_img.png")
        return data

    def get_default_cover_image(self):
        data = pkgutil.get_data("hubstore",
                                "asset/default_cover.png")
        return data



import tkinter as tk
from viewable import Viewable
from suggestion import Suggestion, Engine
from hubstore.misc.constant import SPECIAL_COMMANDS
from hubstore.misc.theme import get_text_search_style


class Toolbar(Viewable):
    def __init__(self, master, views, host):
        super().__init__()
        self._master = master
        self._views = views
        self._host = host
        self._entry_search = None
        self._search_strvar = tk.StringVar()
        self._open_count = tk.StringVar(value="0 Open")
        self._button_all = None
        self._button_favorites = None
        self._button_promoted = None
        self._button_open = None
        self._button_issues = None
        self._suggestion_engine = None

    def wake_search_entry(self):
        self._entry_search.focus_set()

    def update_open_count(self, count):
        text = "{} Open"
        cache = "99+"
        if count < 100:
            cache = str(count)
        text = text.format(cache)
        self._open_count.set(text)

    def update_apps_list(self, apps_list):
        self._suggestion_engine.owners_repos = apps_list

    def _on_search(self):
        search = self._search_strvar.get()
        github_prefix = "https://github.com/"
        if search.startswith(github_prefix):
            search = search.lstrip(github_prefix)
        self._search_strvar.set("")
        if search.startswith("@"):
            cache = search.split(":")
            if len(cache) != 2:
                return
            command, query = cache
        else:
            command = "@search"
            query = search
        self._interpret_query(command, query)

    def _on_click_all(self):
        self._host.show_all_apps()

    def _on_click_favorites(self):
        self._host.show_favorite_apps()

    def _on_click_promoted(self):
        self._host.show_promoted()

    def _on_click_open(self):
        self._host.show_open()

    def _on_click_about(self):
        self._host.show_about()

    def _interpret_query(self, command, query):
        if not query:
            return
        if command == "@from":
            self._host.show_apps_from(query)
            return
        cache = query.split("/")
        if len(cache) != 2:
            return
        owner, repo = cache
        if command == "@run":
            self._host.run_app(owner, repo)
        else:  # search
            self._host.search(owner, repo)

    def _build(self):
        self._body = tk.Frame(self._master)
        # install search field
        self._install_search_field()
        # install buttons
        self._install_buttons()

    def _on_map(self):
        self._link_suggestion()
        self._entry_search.bind("<Return>",
                                lambda e, self=self: self._on_search(), "+")

    def _on_destroy(self):
        pass

    def _install_search_field(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, side=tk.LEFT, expand=1)
        # Search:
        strvar = tk.StringVar(value="Search:")
        entry = tk.Entry(frame, width=7, textvariable=strvar,
                         state="readonly", cursor="hand1")
        entry.pack(side=tk.LEFT)
        entry.bind("<Button-1>", lambda e, self=self: self._on_search())
        get_text_search_style().target(entry)
        # search entry
        self._entry_search = tk.Entry(frame,
                                      textvariable=self._search_strvar)
        self._entry_search.pack(side=tk.LEFT, fill=tk.X,
                                expand=1, padx=(0, 5), anchor="center")

    def _install_buttons(self):
        # button All
        self._button_all = tk.Button(self._body, text="All",
                                     command=self._on_click_all)
        self._button_all.pack(side=tk.LEFT, padx=(0, 5))
        # button Favorites
        self._button_favorites = tk.Button(self._body, text="Favorites",
                                     command=self._on_click_favorites)
        self._button_favorites.pack(side=tk.LEFT, padx=(0, 5))
        # button Promoted
        self._button_promoted = tk.Button(self._body, text="Promoted",
                                          command=self._on_click_promoted)
        self._button_promoted.pack(side=tk.LEFT, padx=(0, 5))
        # button Open
        self._button_open = tk.Button(self._body, textvariable=self._open_count,
                                          command=self._on_click_open)
        self._button_open.pack(side=tk.LEFT, padx=(0, 5))
        # button About
        command = lambda self=self: self._host.show_about()
        button_about = tk.Button(self._body, text="About",
                                 command=self._on_click_about)
        button_about.pack(side=tk.LEFT)

    def _link_suggestion(self):
        suggestion = Suggestion(self._entry_search)
        self._suggestion_engine = SuggestionEngine(suggestion,
                                                   commands=SPECIAL_COMMANDS)
        suggestion.engine = self._suggestion_engine


class SuggestionEngine(Engine):
    def __init__(self, suggestion, commands=None, owners_repos=None):
        self._suggestion = suggestion
        self._owners_repos = None
        self._owners = None
        self._commands = None
        self._prepare_dataset(commands=commands,
                              owners_repos=owners_repos)
        self._setup()

    @property
    def owners_repos(self):
        return self._owners_repos

    @owners_repos.setter
    def owners_repos(self, val):
        self._prepare_dataset(owners_repos=val)

    def process(self, info, callback):
        if info.special_word:
            return
        word = info.word
        if word.startswith("@"):
            cache = self._basic_search(word, self._commands)
            callback(cache)
            return
        if info.cached_word == "@from:":
            cache = self._basic_search(word, self._owners)
            callback(cache)
            return
        cache = []
        for owner, repo in self._owners_repos:
            if word in owner or word in repo:
                cache.append("{}/{}".format(owner, repo))
        callback(cache)

    def _setup(self):
        self._owners_repos = (self._owners_repos if self._owners_repos
                              else ())
        self._owners = (self._owners if self._owners
                        else ())

    def _basic_search(self, word, dataset):
        cache = []
        for item in dataset:
            if word in item:
                cache.append(item)
        return cache

    def _prepare_dataset(self, commands=None, owners_repos=None):
        if commands is not None:
            self._commands = ["@{}: ".format(command)
                              for command in commands]
        if owners_repos is not None:
            self._owners_repos = owners_repos
            self._owners = [owner for owner, repo in owners_repos]


class Dataset:
    def __init__(self):
        self.owners_repos = None
        self.commands = None



import os
import os.path
import uuid
import shutil
import subprocess
import sys
import time
import random
from hubstore.misc import funcs
from tempfile import TemporaryDirectory
from hubstore.misc import constant
from kurl import Kurl
from shared import Store, Jason, valid_store, valid_jason


def get_store():
    """
    Returns the main store datastore if it exists,
    else returns None
    """
    cache = os.path.join(constant.PYRUSTIC_DATA, "hubstore")
    if not valid_jason(os.path.join(cache, "meta.json")):
        return None
    jason = Jason("meta.json", location=cache)
    if not jason.data:
        return None
    path = jason.data.get("hubstore-apps")
    if not path:
        return None
    cache = os.path.join(path, "data")
    if not valid_store(cache):
        return None
    return Store("data", location=path)


def should_init_hubstore():
    if get_store():
        return False
    return True


def init_hubstore(path):
    """
    Initialize Hubstore.

    Param:
        - path: absolute path in which the folder "hubstore-apps"
        will be created

    Return:
        A dict: {"error_code": int, "error": object, "root_dir": str}
        - The "error_code" is one of:
            0 = all right
            1 = failed to create "hubstore-apps" folder
        - "root_dir" is the absolute path to the folder "hubstore-apps".
        Example: init_hubstore("/path/to/use") will return this "root_dir":
            "/path/to/use/hubstore-apps"
        Inside "hubstore-apps", two main folders:
            - apps: to store apps by owner. Meaning that visible folders
            in "apps" are owners, and inside each owner there are a repo
            - data: this is where the library Store stores data.
    """
    error_code = 0
    error = None
    root_dir = None
    data = {"error_code": error_code,
            "error": error,
            "root_dir": root_dir}
    # create hubstore-apps
    error, root_dir = _create_hubstore_apps_folder(path)
    if error is not None:
        data["error_code"] = 1
        data["error"] = error
        return data
    # register Hubstore in PyrusticData
    _register_hubstore_in_pyrustic_data(root_dir)
    data["root_dir"] = root_dir
    return data


def rate(kurl):
    """
    Get Rate Limit
    Return: status_code, status_text, data
    data = {"limit": int, "remaining": int}
    """
    target = "https://api.github.com"
    url = "{}/rate_limit".format(target)
    response = kurl.request(url)
    json = response.json
    status_code, status_text = response.status
    data = {}
    if status_code == 304:
        data = response.cached_response.json
    if (status_code in (200, 304)) and json:
        data["limit"] = json["resources"]["core"]["limit"]
        data["remaining"] = json["resources"]["core"]["remaining"]
    return status_code, status_text, data


def fetch(owner, repo, kurl):
    """ Fetch resource
    Param:
        - owner: str
        - repo: str
        - kurl: Kurl object

    Return:
        status_code, status_text, json_data
        Json_data: {"description": str, "stargazers": int,
                    "tag": str, "published_at": str,
                    "downloads": int, "size": int,
                    "repository": str, "owner_repo": str,
                    "assets": list}
    """
    # fetch repo description
    data = dict()
    data["owner_repo"] = "{}/{}".format(owner, repo)
    data["repository"] = "https://github.com/{}/{}".format(owner, repo)
    status_code, status_text, json = fetch_description(owner, repo, kurl)
    if status_code in (200, 304):
        data["description"] = json["description"]
        data["stargazers"] = json["stargazers_count"]
        status_code, status_text, json = fetch_latest_release(owner, repo, kurl)
        if status_code in (200, 304):
            data["tag"] = json["tag_name"]
            data["published_at"] = funcs.badass_iso_8601_date_parser(json["published_at"])
            data["assets"] = []
            for asset in json["assets"]:
                _, ext = os.path.splitext(asset["name"])
                if ext != ".whl":
                    continue
                cache = {"name": asset["name"],
                         "size": asset["size"],
                         "url": asset["browser_download_url"],
                         "download_count": asset["download_count"],
                         "content_type": asset["content_type"]}
                data["assets"].append(cache)
    return status_code, status_text, data


def useful_info(data):
    """ Return useful info in a dict. Return None if the dict data is empty """
    if not data:
        return None
    info = {"tag_name": data["tag_name"],
            "published_at": funcs.badass_iso_8601_date_parser(data["published_at"]),
            "assets": []}
    for asset in data["assets"]:
        cache = {"name": asset["name"],
                 "size": asset["size"],
                 "url": asset["browser_download_url"],
                 "created_at": funcs.badass_iso_8601_date_parser(asset["created_at"]),
                 "updated_at": funcs.badass_iso_8601_date_parser(asset["updated_at"]),
                 "uploader_login": asset["uploader"]["login"],
                 "download_count": asset["download_count"],
                 "content_type": asset["content_type"]}
        info["assets"].append(cache)
    return info


def download(name, url, kurl):
    """
    Download the resource (url) with kurl object
    Return a boolean is_success, error, and tempdata
    Tempdata: (tempdir, filename)
    """
    is_success = True
    error = None
    response = kurl.request(url)
    data = response.body
    if response.code == 304:
        if response.cached_response:
            data = response.cached_response.body
    if data is None:
        return None, "The server returned an empty body", None
    tempdir = None
    filename = None
    try:
        tempdir = TemporaryDirectory()
        filename = os.path.join(tempdir.name, name)
        with open(filename, "wb") as file:
            file.write(data)
    except Exception as e:
        error = e
        is_success = False
    tempdata = {"tempdir": tempdir, "filename": filename}
    return is_success, error, tempdata


def install(owner, repo, name, tempdata):
    """
    Install the cached wheel file
    Return bool is_success, error
    """
    is_success, error = _install(owner, repo, tempdata)
    if not is_success:
        return False, error
    store = get_store()
    app_pkg = name.split("-")[0]
    apps = store.get("apps")
    if owner not in apps:
        apps[owner] = list()
    if repo not in apps[owner]:
        apps[owner].append(repo)
    apps.save()
    owner_repo = "{}/{}".format(owner, repo)
    timestamp_install = int(time.time())
    if not store.get(owner_repo):
        cache = {"app_pkg": app_pkg, "description": None,
                 "version": None, "small_img": None,
                 "large_img": None, "timestamp_install": timestamp_install,
                 "backup_version": None}
        store.set(owner_repo, cache)
    else:
        data = store.get(owner_repo)
        data["app_pkg"] = app_pkg
        data["timestamp_install"] = timestamp_install
        data.save()
    return True, None


def app_metadata(owner, repo):
    store = get_store()
    owner_repo = "{}/{}".format(owner, repo)
    probed_data = store.get(owner_repo)
    app_pkg = probed_data["app_pkg"]
    #app_pkg = store.get("{}/{}".format(owner, repo))["app_pkg"]
    path = get_path(owner, repo, store=store)
    data = {"version": None, "description": None,
            "small_img": None, "large_img": None}
    metadata_filename = None
    for item in os.listdir(path):
        if item.startswith(app_pkg) and item.endswith("dist-info"):
            metadata_filename = os.path.join(path, item, "METADATA")
            break
    if metadata_filename is None:
        return data
    raw = funcs.dirty_metadata_parser(metadata_filename)
    mapping = {"Version": "version",
               "Summary": "description"}
    for item in raw:
        key, value = item
        if key in mapping:
            data[mapping[key]] = value

    # extract data from $APP_PKG/pyrustic_data/hubstore
    pyrustic_data = os.path.join(path, app_pkg,
                                 "pyrustic_data",
                                 "hubstore")
    jason = Jason("img.json", location=pyrustic_data)
    if jason.data:
        data["small_img"] = jason.data.get("small_img")
        data["large_img"] = jason.data.get("large_img")
    return data


def backup(owner, repo):
    """
    Backup the current version of owner/repo
    Return bool is_success, error
    """
    # variables
    is_success = True
    error = None
    hubstore_apps = get_store().location
    owner_folder = os.path.join(hubstore_apps, "apps", owner)
    repo_folder = os.path.join(owner_folder, repo)
    habitat_folder = os.path.join(repo_folder, "habitat")
    backup_folder = os.path.join(repo_folder, "backup")
    backup_habitat_folder = os.path.join(backup_folder, "habitat")
    # if the repo habitat doesn't exist, return
    if not os.path.exists(habitat_folder):
        return False, None
    # create backup folder if it doesn't exist
    if not os.path.exists(backup_folder):
        try:
            os.makedirs(backup_folder)
        except Exception as e:
            return False, e
    # move habitat backup into trash if it exists
    if os.path.exists(backup_habitat_folder):
        trash = os.path.join(constant.PYRUSTIC_DATA, "trash")
        if not os.path.exists(trash):
            try:
                os.makedirs(trash)
            except Exception as e:
                return False, e
        while True:
            random_data = str(uuid.uuid4().hex)
            cache = os.path.join(trash, random_data)
            if not os.path.exists(cache):
                break
        moveto_is_success, moveto_error = _moveto(backup_habitat_folder, cache)
        if not moveto_is_success:
            return False, moveto_error
    # get app metadata
    # TODO
    # move habitat into backup folder
    src = habitat_folder
    dest = backup_folder
    is_success, error = _moveto(src, dest)
    if is_success:
        # save metadata
        owner_repo = "{}/{}".format(owner, repo)
        store = get_store()
        data = store.get(owner_repo)
        data["backup_version"] = data["version"]
        data.save()
        return True, None
    return is_success, error


def rollback(owner, repo):
    """
    Rollback. Return boolean is_success, error
    """
    # variables
    is_success = True
    error = None
    hubstore_apps = get_store().location
    owner_folder = os.path.join(hubstore_apps, "apps", owner)
    repo_folder = os.path.join(owner_folder, repo)
    habitat_folder = os.path.join(repo_folder, "habitat")
    backup_folder = os.path.join(repo_folder, "backup")
    backup_habitat_folder = os.path.join(backup_folder, "habitat")
    # if the backup habitat doesn't exist, return
    if not os.path.exists(backup_habitat_folder):
        return False, None
    # create backup folder if it doesn't exist
    #if not os.path.exists(backup_folder):
    #    try:
    #        os.makedirs(backup_folder)
    #    except Exception as e:
    #        return False, e
    # move habitat folder into trash if it exists
    if os.path.exists(habitat_folder):
        cache = os.path.join(constant.PYRUSTIC_DATA, "trash")
        if not os.path.exists(cache):
            try:
                os.makedirs(cache)
            except Exception as e:
                return False, e
        while True:
            random_data = str(uuid.uuid4().hex)
            cache = os.path.join(cache, random_data)
            if not os.path.exists(cache):
                break
        cache = _moveto(habitat_folder, cache)
        if not cache[0]:
            return False, cache[1]
    # move backup habitat folder into repo folder
    src = backup_habitat_folder
    dest = repo_folder
    if os.path.exists(src):
        is_success, error = _moveto(src, dest)
    return is_success, error


def uninstall(owner, repo):
    """
    Uninstall an app by providing its owner and repo.
    Return bool is_success, error
    """
    is_success = None
    error = None
    hubstore_apps = get_store().location
    owner_folder = os.path.join(hubstore_apps, "apps", owner)
    repo_folder = os.path.join(owner_folder, repo)
    habitat_folder = os.path.join(repo_folder, "habitat")
    backup_folder = os.path.join(repo_folder, "backup")
    backup_habitat_folder = os.path.join(backup_folder, "habitat")
    # if the repo folder doesn't exist, return
    if not os.path.exists(repo_folder):
        return False, None
    cache = os.path.join(constant.PYRUSTIC_DATA, "trash")
    if not os.path.exists(cache):
        try:
            os.makedirs(cache)
        except Exception as e:
            return False, e
    while True:
        random_data = str(uuid.uuid4().hex)
        cache = os.path.join(cache, random_data)
        if not os.path.exists(cache):
            break
    is_success, error = _moveto(repo_folder, cache)
    if is_success:
        store = get_store()
        apps = store.get("apps")
        if owner in apps:
            for i, item in enumerate(apps[owner].copy()):
                if repo == item:
                    del apps[owner][i]
                    break
            apps.save()
        store.delete("{}/{}".format(owner, repo))
    return is_success, error


def run(owner, repo):
    root_dir = get_path(owner, repo)
    app_pkg = get_app_pkg(owner, repo)
    if (not app_pkg
            or not root_dir
            or not os.path.exists(root_dir)):
        return False, "This app doesn't exist"
    try:

        p = subprocess.Popen([sys.executable, "-m", app_pkg],
                             cwd=root_dir)
        p.communicate()
    except Exception as e:
        return False, e
    else:
        return True, None


def OLDrollback(owner, repo):
    """
    Rollback. Return boolean is_success, error
    """
    hubstore_apps = _hubstore_apps_folder()
    backup_path = os.path.join(hubstore_apps, "backup")
    backup_owner_path = os.path.join(backup_path, owner)
    backup_repo_path = os.path.join(backup_owner_path, repo)
    hubstore_apps_apps_path = os.path.join(hubstore_apps, "apps",)
    hubstore_apps_owner_path = os.path.join(hubstore_apps_apps_path, owner)
    hubstore_apps_repo = os.path.join(hubstore_apps_owner_path,
                                      "{}-habitat".format(repo))
    if not os.path.exists(backup_repo_path) and os.path.exists(hubstore_apps_repo):
        return False, "There aren't any backup for this app"
    elif not os.path.exists(hubstore_apps_repo):
        return False, "This app doesn't exist"
    if not os.path.exists(hubstore_apps_owner_path):
        try:
            os.mkdir(hubstore_apps_owner_path)
        except Exception as e:
            return False, e
    elif os.path.exists(hubstore_apps_repo):
        cache = os.path.join(constant.PYRUSTIC_DATA, "cache")
        if not os.path.exists(cache):
            try:
                os.makedirs(cache)
            except Exception as e:
                return False, e
        while True:
            random_data = str(uuid.uuid4().hex)
            cache = os.path.join(cache, random_data)
            if not os.path.exists(cache):
                break
        is_success, error = _moveto(hubstore_apps_repo, cache)
        if not is_success:
            return False, error
    return _moveto(backup_repo_path, hubstore_apps_repo)


def get_app_pkg(owner, repo, store=None):
    store = store if store else get_store()
    owner_repo = "{}/{}".format(owner, repo)
    data = store.get(owner_repo)
    app_pkg = data["app_pkg"]
    return app_pkg


def get_path(owner, repo, store=None):
    """
    Use this to get the path to the app owner/repo
    Return a path string or None
    """
    if not store:
        store = get_store()
    if not store:
        return None
    path = os.path.join(store.location,
                        "apps", owner, repo,
                        "habitat")
    if os.path.exists(path):
        return path
    return None


def parse_owner_repo(val):
    data_splitted = val.split("/")
    if len(data_splitted) != 2:
        return None, None
    if data_splitted[0] == "":
        data_splitted[0] = data_splitted[1]
    elif data_splitted[1] == "":
        data_splitted[1] = data_splitted[0]
    owner, repo = data_splitted
    return owner, repo


def get_kurl():
    """
    Generate a Kurl object
    """
    headers = {"Accept": "application/vnd.github.v3+json",
               "User-Agent": "Pyrustic"}
    return Kurl(headers=headers)


def fetch_latest_release(owner, repo, kurl):
    target = "https://api.github.com"
    url = "{}/repos/{}/{}/releases/latest".format(target, owner, repo)
    response = kurl.request(url)
    json = response.json
    status_code, status_text = response.status
    if status_code == 304:
        json = response.cached_response.json
    return status_code, status_text, json


def fetch_description(owner, repo, kurl):
    target = "https://api.github.com"
    url = "{}/repos/{}/{}".format(target, owner, repo)
    response = kurl.request(url)
    json = response.json
    status_code, status_text = response.status
    if status_code == 304:
        json = response.cached_response.json
    return status_code, status_text, json


def normpath(target, path):
    path = path.lstrip("./\\")
    path = path.rstrip("/\\")
    cache = ""
    parts = []
    for char in path:
        if char in ("/", "\\"):
            parts.append(cache)
            cache = ""
            continue
        cache += char
    parts.append(cache)
    return os.path.join(target, *parts)


def _create_hubstore_apps_folder(parent_path):
    """
    parent_path is the absolute path in which the folder "hubstore-apps"
    will be created.
    return (error, root_dir)
    """
    hubstore_apps = os.path.join(parent_path, "hubstore-apps")
    apps_dir = os.path.join(hubstore_apps, "apps")
    for item in (hubstore_apps, apps_dir):
        if not os.path.exists(item):
            try:
                os.mkdir(item)
            except Exception as e:
                return e, hubstore_apps
    # initialize Store
    store = Store("data", location=hubstore_apps)
    if store.new:
        store.set("apps", dict())
        store.set("favorites", list())
    return None, hubstore_apps


def _register_hubstore_in_pyrustic_data(hubstore_apps):
    location = os.path.join(constant.PYRUSTIC_DATA, "hubstore")
    jason = Jason("meta.json", location=location)
    jason.data = {"hubstore-apps": hubstore_apps}
    jason.save()


def _install(owner, repo, tempdata):
    """
    Install the cached wheel file
    Return bool is_success, error
    """
    hubstore_apps = get_store().location
    tempdir = tempdata["tempdir"]
    filename = tempdata["filename"]
    src = filename
    dest = os.path.join(hubstore_apps, "apps",
                        owner, repo, "habitat")
    if not os.path.exists(dest):
        os.makedirs(dest)
    try:
        args = ["-m", "pip", "install",
                "--upgrade",
                "--upgrade-strategy", "eager",
                "--target={}".format(dest), src]
        p = subprocess.Popen([sys.executable, *args],
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE)
        out, err = p.communicate()
        if err:
            # pip outputs a warning error about its own version
            # return False, err
            pass
    except Exception as e:
        return False, e
    finally:
        tempdir.cleanup()
    if not os.listdir(dest):
        return False, "Unknown error"
    return True, None


def _moveto(src, dest):
    """
    If the DEST exists:
        * Before moveto *
        - /home/lake (SRC)
        - /home/lake/fish.txt
        - /home/ocean (DEST)
        * Moveto *
        moveto("/home/lake", "/home/ocean")
        * After Moveto *
        - /home/ocean
        - /home/ocean/lake
        - /home/ocean/lake/fish.txt
    Else IF the DEST doesn't exist:
        * Before moveto *
        - /home/lake (SRC)
        - /home/lake/fish.txt
        * Moveto *
        moveto("/home/lake", "/home/ocean")
        * After Moveto *
        - /home/ocean
        - /home/ocean/fish.txt


    Move a file or directory (src) to a destination folder (dest)
    """
    if not os.path.exists(src):
        return False, None
    try:
        shutil.move(src, dest)
    except Exception as e:
        return False, e
    return True, None



import tkinter as tk
from viewable import Viewable
from megawidget.scrollbox import ScrollBox
from megawidget.toast import Toast
from hubstore.misc.theme import get_highlight_style, get_unhighlight_style


MAX_TILES_BY_ROW = 4


class Pane(Viewable):
    def __init__(self, parent_view):
        super().__init__()
        self._master = parent_view.body
        self._views = parent_view.views
        self._host = parent_view.host
        self._strvar_title = tk.StringVar()
        self._strvar_notification = tk.StringVar(value="")
        self._frame_matrix = None
        self._frame_row_cache = None  # will hold: [tk_frame, int_children_count]
        self._image_cache = {}
        self._image_cachex = None

    def populate(self, title, data):
        self._strvar_title.set(title)
        self._clear_frame_matrix()
        if not data:
            Toast(self._body, message="No data available")
            return
        self._frame_row_cache = None
        self._image_cache = {}
        self._loop(data, 0)

    def set_notification(self, text=None):
        text = "" if not text else text
        self._strvar_notification.set(text)

    def _build(self):
        self._body = tk.Frame(self._master)
        self._scrollbox = ScrollBox(self._body)
        self._scrollbox.pack(expand=1, fill=tk.BOTH)
        frame_header = tk.Frame(self._scrollbox.box)
        frame_header.pack(fill=tk.X, pady=(10, 1))
        self._frame_matrix = tk.Frame(self._scrollbox.box)
        self._frame_matrix.pack(fill=tk.BOTH, expand=1)
        # label line
        #canvas = tk.Canvas(frame_header, width=43, height=10,
        #                   highlightthickness=0, borderwidth=0)
        #canvas.pack(side=tk.LEFT, padx=0, pady=0)
        #with open("/home/alex/line.png", "rb") as file:
        #    data = file.read()
        #image = tk.PhotoImage(data=data)
        #self._image_cachex = image
        #canvas.create_image(0, 0, image=image, anchor="nw")
        #frame_line = tk.Frame(frame_header, name="frame_line", width=25,
        #                      height=1, pady=0)
        #frame_line.pack(side=tk.LEFT, padx=(2, 3))
        # label title
        label_title = tk.Label(frame_header, name="label_title",
                               textvariable=self._strvar_title)
        label_title.pack(side=tk.LEFT, padx=12)
        # label notification
        label_notification = tk.Label(frame_header, name="label_notification",
                                      textvariable=self._strvar_notification)
        label_notification.pack(side=tk.RIGHT)

    def _on_map(self):
        pass

    def _on_destroy(self):
        pass

    def _loop(self, data, index):
        if index == len(data):
            return
        owner, repo = data[index]
        self._populate(owner, repo)
        index += 1
        command = (lambda self=self, data=data, index=index:
                    self._loop(data, index))
        self._body.after(0, command)

    def _populate(self, owner, repo):
        if self._frame_row_cache is None:
            self._add_frame_row()
        elif self._frame_row_cache[1] == MAX_TILES_BY_ROW:
            self._add_frame_row()
        frame_row = self._frame_row_cache[0]
        children_count = self._frame_row_cache[1]
        self._add_tile(frame_row, owner, repo)
        self._frame_row_cache[1] = children_count + 1

    def _clear_frame_matrix(self):
        if self._frame_matrix:
            self._frame_matrix.destroy()
        self._frame_matrix = tk.Frame(self._scrollbox.box)
        self._frame_matrix.pack(fill=tk.BOTH, expand=1)

    def _add_frame_row(self):
        frame = tk.Frame(self._frame_matrix)
        frame.pack(fill=tk.X, padx=5)
        self._frame_row_cache = [frame, 0]

    def _add_tile(self, frame_row, owner, repo):
        tile = tk.Frame(frame_row)
        tile.pack(side=tk.LEFT, padx=5, pady=(0, 10))
        #tile.config(highlightthickness=2)
        #tile.config(highlightbackground="white")
        get_unhighlight_style().target(tile)
        # owner name
        entry_owner = tk.Entry(tile,borderwidth=0,width=0,
                               name="entry_owner_name",
                               cursor="hand1")
        entry_owner.pack(anchor="center", fill=tk.X,
                         padx=2, pady=(2, 1))
        entry_owner.insert(0, owner)
        entry_owner.config(state="readonly", cursor="hand1")
        # canvas
        canvas = tk.Canvas(tile, width=200, height=80,
                           highlightthickness=0, borderwidth=0,
                           cursor="hand1")
        canvas.pack(padx=2, pady=1, fill=tk.BOTH, expand=1)
        self._set_image(canvas, owner, repo)
        # repo name
        entry_repo = tk.Entry(tile, borderwidth=0, width=0,
                              name="entry_repo_name",
                              cursor="hand1")
        entry_repo.pack(anchor="w", fill=tk.X,
                        padx=2, pady=(0, 2), ipady=2)
        entry_repo.insert(0, repo)
        entry_repo.config(state="readonly")
        # binding
        self._bind_handler_owner(owner, tile, entry_owner)
        self._bind_handler_canvas(owner, repo, tile, canvas)
        self._bind_handler_repo(owner, repo, tile, entry_repo)

    def _set_image(self, canvas, owner, repo):
        self.body.update_idletasks()
        data = self._host.data.get_image(owner, repo)
        if not data:
            return
        image = tk.PhotoImage(data=data)
        #image.zoom(25)
        #image = image.subsample(3, 3)
        self._image_cache["{}/{}".format(owner, repo)] = image
        canvas.create_image(0, 0, image=image, anchor="nw")

    def _bind_handler_owner(self, owner, tile, entry_owner):
        # on click
        command = (lambda event, self=self, owner=owner:
                   self._on_click_owner(owner))
        entry_owner.bind("<Button-1>", command, "+")
        # on enter
        command = (lambda event, tile=tile:
                   get_unhighlight_style().target(tile))
        entry_owner.bind("<Enter>", command, "+")
        # on leave
        entry_owner.bind("<Leave>", command, "+")

    def _bind_handler_canvas(self, owner, repo, tile, canvas):
        # handle Click on Canvas
        command = (lambda event, self=self, owner=owner,
                          repo=repo:
                   self._on_click_canvas(owner, repo))
        canvas.bind("<Button-1>", command)
        # handle on enter Canvas
        command = (lambda event, tile=tile:
                   get_highlight_style().target(tile))
        canvas.bind("<Enter>", command, "+")
        # handle on leave Canvas
        command = (lambda event, tile=tile:
                   get_unhighlight_style().target(tile))
        canvas.bind("<Leave>", command, "+")

    def _bind_handler_repo(self, owner, repo, tile, entry_repo):
        command_entry = (lambda event, self=self, owner=owner, repo=repo:
                         self._on_click_repo(owner, repo))
        # on click
        entry_repo.bind("<Button-1>", command_entry)
        # on enter
        command = (lambda event, tile=tile:
                   get_unhighlight_style().target(tile))
        entry_repo.bind("<Enter>", command, "+")
        # on leave
        entry_repo.bind("<Leave>", command, "+")

    def _on_click_owner(self, owner):
        self._host.show_apps_from(owner)

    def _on_click_canvas(self, owner, repo):
        self._host.run_app(owner, repo)

    def _on_click_repo(self, owner, repo):
        self._host.show_app_info(owner, repo)



import tkinter as tk
from viewable import Viewable
from megawidget.scrollbox import ScrollBox
from megawidget.toast import Toast
from hubstore.misc.theme import get_entry_description_style, \
    get_info_entry_owner_repo_style, get_package_size_style
from cyberpunk_theme.widget.button import get_button_dark_style, get_button_blue_filled_style, get_button_yellow_style


class Promoted(Viewable):

    def __init__(self, parent_view, data=None):
        super().__init__()
        self._data = data
        self._master = parent_view.body
        self._views = parent_view.views
        self._host = parent_view.host
        self._image_cache = None
        self._scrollbox = None

    def _build(self):
        self._body = tk.Toplevel(name="promoted_toplevel")
        self._body.geometry("380x250")
        self._body.title("Promoted Apps")
        self._body.resizable(False, False)
        #self._body.overrideredirect(1)
        self._scrollbox = ScrollBox(self._body)
        self._scrollbox.pack(expand=1, fill=tk.BOTH)
        # install header
        #self._install_header()
        # install central
        self._install_central()
        # install footer
        self._install_footer()

    def _on_map(self):
        super()._on_map()
        if not self._data:
            self._body.destroy()
            Toast(message="Nothing to show !",
                  duration=1000)

    def _on_destroy(self):
        pass

    def _install_header(self):
        """
        # install image
        canvas = tk.Canvas(self._body, width=400, height=100,
                           highlightthickness=0, borderwidth=0,
                           bg="#121519")
        canvas.pack(fill=tk.BOTH, pady=(0,5))
        with open("/home/alex/default_info_cover.png",
                  "rb") as file:
            data = file.read()
        image = tk.PhotoImage(data=data)
        self._image_cache = image
        canvas.create_image(0, 0, image=image, anchor="nw")
        """
        # description
        description = tk.Label(self._scrollbox.box, font=("Liberation Mono", 11, "bold"),
                               text="These are promoted apps")
        description.pack(pady=(10, 10), anchor="center")
        #get_entry_description_style().target(description)

    def _install_central(self):
        # install
        for item in self._data:
            owner_repo = item["repository"]
            description = item["description"]
            github_prefix = "https://github.com/"
            if owner_repo.startswith(github_prefix):
                owner_repo = owner_repo.lstrip(github_prefix)
            self._install_row(owner_repo, description)

    def _install_row(self, owner_repo, description):
        # frame
        frame = tk.Frame(self._scrollbox.box)
        frame.pack(fill=tk.X, padx=(3, 5), pady=10)
        # description
        strvar_description = tk.StringVar(value=description)
        description = tk.Entry(frame,
                               textvariable=strvar_description,
                               state="readonly")
        description.pack(fill=tk.X, padx=2)
        get_entry_description_style().target(description)
        # subframe
        subframe = tk.Frame(frame)
        subframe.pack(fill=tk.X)
        # entry
        strvar_owner_repo = tk.StringVar(value=owner_repo)
        entry = tk.Entry(subframe, textvariable=strvar_owner_repo,
                         state="readonly")
        entry.pack(side=tk.LEFT, fill=tk.X, expand=1)
        get_info_entry_owner_repo_style().target(entry)
        # button get
        command = (lambda self=self, owner_repo=owner_repo:
                   self._on_click_get(owner_repo))
        button_get = tk.Button(subframe, text="Get",
                           command=command)
        button_get.pack(side=tk.LEFT, padx=(5, 0))
        # button visit
        command = (lambda self=self, owner_repo=owner_repo:
                   self._on_click_visit(owner_repo))
        button_visit = tk.Button(subframe, text="Visit",
                               command=command)
        button_visit.pack(side=tk.LEFT, padx=(5, 0))

    def _install_footer(self):
        frame = tk.Frame(self._body)
        frame.pack(fill=tk.X, padx=3, pady=(15, 3))
        # button close
        button_more = tk.Button(frame, text="More",
                           command=self._on_click_more)
        button_more.pack(side=tk.RIGHT, padx=(3, 0))
        # button more
        button_close = tk.Button(frame, text="Close",
                           command=self.destroy)
        button_close.pack(side=tk.RIGHT)
        get_button_blue_filled_style().target(button_more)

    def _on_click_get(self, owner_repo):
        cache = owner_repo.split("/")
        if len(cache) != 2:
            return
        owner, repo = cache
        self._host.search(owner, repo)
        self.destroy()

    def _on_click_visit(self, owner_repo):
        github_prefix = "https://github.com/"
        repository = "{}{}".format(github_prefix, owner_repo)
        self._host.open_website(repository)


    def _on_click_more(self):
        self.destroy()
        self._host.show_promoted()
